"""
Implements the Behave Restful application object which initializes the global
execution context providing access to directories and other services to 
simplify testing.  
"""
import os

import requests

import behave_restful._definitions as _defs
import behave_restful._hooks as _hooks

from behave_restful._hooks import (
    BEFORE_ALL, 
    AFTER_ALL, 
    BEFORE_FEATURE, 
    AFTER_FEATURE, 
    BEFORE_SCENARIO, 
    AFTER_SCENARIO,
    BEFORE_STEP,
    AFTER_STEP,
    BEFORE_TAG,
    AFTER_TAG
)

class BehaveRestfulApp(object):
    """
    Behave Restful application class used to initialize the execution context
    for a test run.
    """

    def initialize_context(self, context, test_dir):
        """
        Initializes the provide context object exposing through it relevant
        directories and services. 

        :param context:
            A behave context object.
        :param test_dir:
            Directory where the tests are located. This is, usually, the 
            features folder of the test project.
        """
        context.test_dir = test_dir
        context.working_dir = os.getcwd()
        context.default_session = requests
        context.session = context.default_session
        _defs.DefinitionInitializer().initialize(context)
        _hooks.EnvironmentHookInitializer().initialize(context)


    

    


"""
Provides properties with information about ``behave_restful``.
"""

project = 'behave-restful'
description = 'Implements Gherking language for REST services.'
long_description = '''
Behave Restful implements a language based on Gherkin that allows to test
REST services using the behave python tool.
'''
copyright = '2017 Behave Restful'
author = 'Isaac Rodriguez'
author_email = 'isaac_rodriguez@live.com'
maintainer = 'Isaac Rodriguez'

version = '0.1'
release = '0.1.4'

classifiers = [
    'Development Status :: 1 - Planning',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python',
]

keywords = 'behave gherking rest test bdd'


"""
"""

def send_get(context):
    """
    """
    params = _get_params(context)
    context.response = context.session.get(
        context.request_url,
        params=params
    )


def send_post(context):
    """
    """
    params = _get_params(context)
    context.response = context.session.post(
        context.request_url,
        params=params,
        json=context.request_json_payload
    )


def send_put(context):
    """
    """
    params = _get_params(context)
    context.response = context.session.put(
        context.request_url,
        params=params,
        json=context.request_json_payload
    )


def send_delete(context):
    """
    """
    params = _get_params(context)
    context.response = context.session.delete(
        context.request_url,
        params=params
    )



def _get_params(context):
    return context.request_params if hasattr(context, 'request_params') else None


"""
"""
import json

from assertpy import assert_that, fail
import jsonpath_rw as jp
import jsonschema

from behave_restful.xpy import HTTPStatus

def response_status_is(response, expected_status):
    """
    """
    expected_status = _as_numeric_status(expected_status)
    actual_status = response.status_code
    assert_that(actual_status).is_equal_to(expected_status)


def response_json_matches(response, schema_str):
    """
    """
    schema = json.loads(schema_str)
    json_body = response.json()
    jsonschema.validate(json_body, schema)


def response_json_matches_at(response, json_path, expected_str):
    """
    """
    values = _get_values(response.json(), json_path)
    [assert_that(actual_value).is_equal_to(eval(expected_str)) for actual_value in values]
    


def _as_numeric_status(status):
    status = status.replace(' ', '_')
    numeric_status = getattr(HTTPStatus, status.upper(), None)
    if not numeric_status:
        numeric_status = int(status)
    return numeric_status


def _get_values(json_body, json_path):
    results = jp.parse(json_path).find(json_body)
    if not results: fail('Match not found at <{path}> for <{body}>'.format(path=json_path, body=json_body))
    values = [result.value for result in results]
    return values    




"""
This module implements a bolt task that allows to execute feature tests using
behave. This task works with any behave project and also with Behave Restful
projects.
"""
import os.path

import behave.__main__ as behave_main
import bolt.errors as bolt_errors


class RunBehaveRestfulTask(object):
    """
    Bolt task that allows executing Behave Restful through bolt.
    """

    def __call__(self, **kwargs):
        self.config = kwargs.get('config')
        self._configure()
        self._execute()


    def _configure(self):
        self.features_dir = self.config.get('directory')
        if not self.features_dir:
            raise FeaturesDirectoryNotSpecifiedError()
        if not self._exists(self.features_dir): 
            raise FeaturesDirectoryDoesNotExistError(self.features_dir)
        self.definition = self.config.get('definition')
        self.options = self.config.get('options') or {}


    def _execute(self):
        options_parser = BehaveOptionsParser()
        arguments = options_parser.parse(self.options)
        if self.definition:
            arguments.extend(['-D', 'definition={d}'.format(d=self.definition)])
        arguments.append(self.features_dir)
        result = self._invoke_behave(arguments)
        if result != 0: raise FeatureTestsFailedError()


    def _exists(self, path):
        return os.path.exists(path)


    def _invoke_behave(self, arguments):
        return behave_main.main(arguments)


def register_tasks(registry):
    registry.register_task('behave-restful', RunBehaveRestfulTask())


OPTION_PREFIX = '--{o}'
NEGATED_OPTION_PREFIX = '--no-{o}'

SHOW_OPTIONS = (
    'show-source',
    'show-timings', 
    'show-skipped'
)
SHOW_PREFIX = 'show'
NEGATED_SHOW_PREFIX = 'no'

DEFINED_DATA_OPTION = 'define'
DEFINED_DATA_FORMAT = '{k}={v}'

TAGS_OPTION = 'tags'
TAGS_DELIMITER = ','

class BehaveOptionsParser(object):
    """
    Parser class that knows how to convert options specified as a dictionary
    to a list of arguments that behave will understand.
    """

    def parse(self, options):
        """
        Parses the specified options returning a list of arguments that can be
        used to invoke behave.

        :param options:
            Dictionary containing the options to be used as arguments to behave.
        """
        parsed = []
        [parsed.extend(self._parse_option(option, value)) for option, value in options.items()]
        return parsed


    def _parse_option(self, option, value):
        formatting_function = self._get_formatting_function(option, value)
        return formatting_function(option, value)


    def _get_formatting_function(self, option, value):
        if self._is_bool(value):
            return self._format_bool
        if self._is_defined_data(option):
            return self._format_defined_data
        if self._is_tags_option(option):
            return self._format_tags
        return self._format_standard
        


    def _format_standard(self, option, value):
        prefixed_option = OPTION_PREFIX.format(o=option)
        return [prefixed_option, value]


    def _is_bool(self, value):
        return isinstance(value, bool)


    def _format_bool(self, option, value):
        if self._is_special_bool(option):
            prefixed_bool = self._format_special_bool(option, value)
        else:
            prefix = OPTION_PREFIX if value else NEGATED_OPTION_PREFIX
            prefixed_bool = prefix.format(o=option)
        return [prefixed_bool]


    def _is_special_bool(self, option):
        return option in SHOW_OPTIONS


    def _format_special_bool(self, option, value):
        if value:
            return OPTION_PREFIX.format(o=option)
        else:
            option = option.replace(SHOW_PREFIX, NEGATED_SHOW_PREFIX)
            return OPTION_PREFIX.format(o=option)


    def _is_defined_data(self, option):
        return option == DEFINED_DATA_OPTION


    def _format_defined_data(self, option, value):
        prefixed_option = OPTION_PREFIX.format(o=option)
        result = []
        for var, val in value.items():
            formatted_value = DEFINED_DATA_FORMAT.format(k=var, v=val)
            result.append(prefixed_option)
            result.append(formatted_value)
        return result


    def _is_tags_option(self, option):
        return option == TAGS_OPTION


    def _format_tags(self, option, value):
        formatted_result = []
        if self._has_multiple_tags_specified(value):
            [formatted_result.extend(entry) for entry in self._format_all_tags(option, value)]
        else:
            formatted_result.extend(self._format_one_tag(option, value))
        return formatted_result


    def _has_multiple_tags_specified(self, value):
        first_item = value[0]
        return isinstance(first_item, list)


    def _format_all_tags(self, option, entries):
        for tag_entry in entries:
            yield self._format_tags(option, tag_entry)


    def _format_one_tag(self, option, value):
        prefixed_option = OPTION_PREFIX.format(o=option)
        formatted_tags = TAGS_DELIMITER.join(value)
        return [prefixed_option, formatted_tags]




class FeaturesDirectoryNotSpecifiedError(bolt_errors.RequiredParameterMissingError):
    """
    Exception class raised when the directory with the feature tests is not
    specified.
    """
    def __init__(self):
        super(FeaturesDirectoryNotSpecifiedError, self).__init__('directory')



class FeaturesDirectoryDoesNotExistError(bolt_errors.ConfigurationValueError):
    """
    Exception class raised when the specified features directory does not
    exist.
    """
    def __init__(self, specified_directory):
        super(FeaturesDirectoryDoesNotExistError, self).__init__('directory', specified_directory)


class FeatureTestsFailedError(bolt_errors.TaskError):
    """
    Raised when behave fails.
    """
    pass



"""
"""
from behave import *

import behave_restful._lang_imp.response_validator as _validate

@then('the response status is {status}')
def step_impl(context, status):
    _validate.response_status_is(context.response, status)


@then('the response json matches')
def step_impl(context):
    schema = context.vars.resolve(context.text)
    _validate.response_json_matches(context.response, schema)


@then('the response json at {json_path} is equal to {value_str}')
def step_impl(context, json_path, value_str):
    json_path = context.vars.resolve(json_path)
    value_str = context.vars.resolve(value_str)
    _validate.response_json_matches_at(context.response, json_path, value_str)



"""
Setup for behave-restful
"""
from setuptools import setup, find_packages
import behave_restful.about as about

packages = find_packages()
requirements = [
    "behave>=1.2.5",
    "jsonpath-rw>=1.4.0",
    "jsonschema>=2.6.0",
    "requests>=2.18.4"
]


setup(
    name=about.project,
    version=about.release,
    author=about.author,
    author_email=about.author_email,
    packages=packages,
    install_requires=requirements
)
